use anyhow::Result;
use blocking::multipart;
use clap::{App, Arg};
use multipart::Part;
use rayon::prelude::*;
use reqwest::{
    blocking::{self, Client},
    header, redirect,
};
use serde::Serialize;
use std::collections::HashMap;
use std::{
    ffi::OsString,
    fs::DirEntry,
    io::{Cursor, Read},
    path::PathBuf,
    sync::{Arc, Mutex},
    thread,
    time::Duration,
};

#[derive(Debug, Serialize)]
struct FileStat {
    st_dev: u64,
    st_ino: u64,
    st_nlink: u64,
    st_mode: u32,
    st_uid: u32,
    st_gid: u32,
    st_rdev: u64,
    st_size: i64,
    st_blksize: i64,
    st_blocks: i64,
    st_atime: i64,
    st_atime_nsec: i64,
    st_mtime: i64,
    st_mtime_nsec: i64,
    st_ctime: i64,
    st_ctime_nsec: i64,
}

impl From<nix::sys::stat::FileStat> for FileStat {
    fn from(x: nix::sys::stat::FileStat) -> Self {
        Self {
            st_dev: x.st_dev,
            st_ino: x.st_ino,
            st_nlink: x.st_nlink,
            st_mode: x.st_mode,
            st_uid: x.st_uid,
            st_gid: x.st_gid,
            st_rdev: x.st_rdev,
            st_size: x.st_size,
            st_blksize: x.st_blksize,
            st_blocks: x.st_blocks,
            st_atime: x.st_atime,
            st_atime_nsec: x.st_atime_nsec,
            st_mtime: x.st_mtime,
            st_mtime_nsec: x.st_mtime_nsec,
            st_ctime: x.st_ctime,
            st_ctime_nsec: x.st_ctime_nsec,
        }
    }
}

#[derive(Default, Debug, Serialize)]
struct Tree {
    childs: HashMap<OsString, Tree>,
    stat: Option<FileStat>,
    xattrs: HashMap<OsString, Option<Vec<u8>>>,
    posix_acls: Option<String>,
    ignored: bool,
    symlink_target: Option<OsString>,
    md5: Option<String>,
    sha1: Option<String>,
    sha256: Option<String>,
    sha512: Option<String>,
}

#[derive(Serialize)]
struct Snapshot {
    disk_partitions: Vec<psutil::disk::Partition>,
    fs_tree: Tree,
}

fn construct_fs_tree(
    cur_tree: Option<Tree>,
    path: &PathBuf,
    dev_whitelist: &Vec<u64>,
    ignored_dirs: &Vec<PathBuf>,
) -> Result<Tree> {
    let mut cur_tree = match cur_tree {
        Some(cur_tree) => cur_tree,
        None => Tree {
            stat: Some(nix::sys::stat::lstat(path)?.into()),
            ..Tree::default()
        },
    };

    if !dev_whitelist.iter().any(|x| match &cur_tree.stat {
        Some(stat) if stat.st_dev == *x => true,
        _ => false,
    }) {
        return Ok(cur_tree);
    }

    if ignored_dirs.iter().any(|x| path.starts_with(x)) {
        cur_tree.ignored = true;
        return Ok(cur_tree);
    }

    let entries: Vec<Result<DirEntry, _>> = match std::fs::read_dir(&path) {
        Ok(x) => x,
        _ => return Ok(cur_tree),
    }
    .collect();

    let cur_tree = {
        let cur_tree = Arc::new(Mutex::new(cur_tree));

        entries.par_iter().for_each(|entry| match entry {
            Ok(entry) => {
                let mut tree = Tree {
                    stat: match nix::sys::stat::lstat(&entry.path()) {
                        Ok(s) => Some(s.into()),
                        _ => None,
                    },
                    posix_acls: match posix_acl::PosixACL::read_acl(&entry.path()) {
                        Ok(acl) => Some(acl.as_text()),
                        _ => None,
                    },
                    xattrs: match xattr::list(&entry.path()) {
                        Ok(xattrs) => xattrs
                            .filter_map(|name| match xattr::get(&entry.path(), &name) {
                                Ok(xattr) => Some((name, xattr)),
                                _ => None,
                            })
                            .collect(),
                        _ => HashMap::new(),
                    },
                    ..Tree::default()
                };

                match entry.file_type() {
                    Ok(file_type) if file_type.is_dir() => {
                        tree = construct_fs_tree(
                            Some(tree),
                            &entry.path(),
                            dev_whitelist,
                            ignored_dirs,
                        )
                        .unwrap();
                    }
                    Ok(file_type) if file_type.is_file() => {
                        if let Ok(mut file) = std::fs::File::open(&entry.path()) {
                            use md5::{Digest, Md5};
                            use sha1::Sha1;
                            use sha2::{Sha256, Sha512};

                            let mut md5 = Md5::new();
                            let mut sha1 = Sha1::new();
                            let mut sha256 = Sha256::new();
                            let mut sha512 = Sha512::new();

                            let buf: &mut [u8] = &mut [0; 8 * 1024];
                            loop {
                                match file.read(buf) {
                                    Ok(0) => {
                                        tree.md5 = Some(hex::encode(md5.finalize()));
                                        tree.sha1 = Some(hex::encode(sha1.finalize()));
                                        tree.sha256 = Some(hex::encode(sha256.finalize()));
                                        tree.sha512 = Some(hex::encode(sha512.finalize()));
                                        break;
                                    }
                                    Ok(n) => {
                                        md5.update(&buf[0..n - 1]);
                                        sha1.update(&buf[0..n - 1]);
                                        sha256.update(&buf[0..n - 1]);
                                        sha512.update(&buf[0..n - 1]);
                                    }
                                    Err(e) if e.kind() == std::io::ErrorKind::Interrupted => {
                                        continue;
                                    }
                                    Err(e) => {
                                        eprintln!("{:#?}", e);
                                        break;
                                    }
                                };
                            }
                        }
                    }
                    Ok(file_type) if file_type.is_symlink() => {
                        tree.symlink_target = std::fs::read_link(&entry.path())
                            .map(|x| x.as_os_str().into())
                            .ok();
                    }
                    _ => {}
                };

                {
                    let mut locked = cur_tree.lock().unwrap();
                    locked.childs.insert(entry.file_name(), tree);
                }
            }
            _ => {}
        });

        Arc::try_unwrap(cur_tree).unwrap().into_inner().unwrap()
    };

    Ok(cur_tree)
}

fn upload_to_erp5<R: Read + Send + 'static>(
    reader: R,
    size: u64,
    base_url: &str,
    username: &str,
    password: &str,
    reference: &str,
) -> Result<()> {
    let mut headers = header::HeaderMap::new();
    headers.insert(
        header::AUTHORIZATION,
        header::HeaderValue::from_str(&format!(
            "Basic {}",
            &base64::encode(&format!("{}:{}", username, password))
        ))?,
    );

    let client = Client::builder()
        .redirect(redirect::Policy::none())
        .default_headers(headers)
        .build()?;

    let resp = client
        .get(&format!("{}/portal_contributions/newContent", base_url))
        .query(&[
            ("portal_type", "File"),
            ("filename", &format!("{}.metadata", &reference)),
            ("container_path", "document_module"),
            ("data", ""),
        ])
        .send()?;

    // workaround for ERP5 disappearing documents race condition bug
    thread::sleep(Duration::from_secs(3));

    let doc_url: reqwest::Url = resp
        .headers()
        .get("X-Location")
        .ok_or_else(|| anyhow::format_err!("X-Location must exist"))?
        .to_str()?
        .parse()?;

    let _resp = client
        .post(doc_url.to_owned())
        .multipart(
            multipart::Form::new()
                .text("form_id", "File_view")
                .text("object_path", doc_url.path().to_owned())
                .text("field_my_reference", reference.to_owned())
                .part(
                    "field_my_file",
                    Part::reader_with_length(reader, size)
                        .file_name(format!("{}.metadata", &reference)),
                )
                .text("Base_edit:method", ""),
        )
        .send()?;

    // workaround for ERP5 disappearing documents race condition bug
    thread::sleep(Duration::from_secs(3));

    let mut params = HashMap::new();

    params.insert("dialog_id", "Base_viewWorkflowActionDialog");
    params.insert("dialog_method", "Workflow_statusModify");
    params.insert("form_id", "File_view");
    params.insert("object_path", doc_url.path());
    params.insert("field_your_comment", "");
    params.insert("field_your_workflow_action", "share_action");
    params.insert("Base_callDialogMethod:method", "");

    let _resp = client.post(doc_url.to_owned()).form(&params).send()?;

    // workaround for ERP5 disappearing documents race condition bug
    thread::sleep(Duration::from_secs(3));

    Ok(())
}

fn main() -> Result<()> {
    let m = App::new("metadata-collect-agent")
        .args(&[
            Arg::with_name("start_directory")
                .required(true)
                .takes_value(true)
                .multiple(false),
            Arg::with_name("erp5-user")
                .takes_value(true)
                .multiple(false)
                .required(true),
            Arg::with_name("erp5-pass")
                .takes_value(true)
                .multiple(false)
                .required(true),
            Arg::with_name("erp5-file-reference")
                .takes_value(true)
                .multiple(false)
                .required(true),
            Arg::with_name("erp5-base-url")
                .takes_value(true)
                .multiple(false)
                .required(true),
            Arg::with_name("ignored-dirs")
                .takes_value(true)
                .multiple(true)
                .required(false),
        ])
        .get_matches();

    let ignored_dirs = m
        .values_of("ignored-dirs")
        .unwrap_or(clap::Values::default())
        .map(PathBuf::from)
        .collect();

    let disk_partitions = psutil::disk::partitions_physical()?;

    let dev_whitelist = disk_partitions
        .iter()
        .map(|p| nix::sys::stat::lstat(p.mountpoint()).unwrap().st_dev)
        .collect();

    let snapshot = Snapshot {
        disk_partitions,
        fs_tree: construct_fs_tree(
            None,
            &PathBuf::from(m.value_of("start_directory").unwrap()),
            &dev_whitelist,
            &ignored_dirs,
        )?,
    };

    let packed = rmp_serde::encode::to_vec_named(&snapshot)?;
    let packed_size = packed.len() as u64;

    upload_to_erp5(
        Cursor::new(packed),
        packed_size,
        m.value_of("erp5-base-url").unwrap(),
        m.value_of("erp5-user").unwrap(),
        m.value_of("erp5-pass").unwrap(),
        m.value_of("erp5-file-reference").unwrap(),
    )?;

    Ok(())
}
